# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from selenium.webdriver.common.by import By


class ProxyDescriptor:
    """Describes an Addon and an Action to be executed via the Agent

        Args:
            guid (str): a unique GUID used to identify the addon
            classname (str): the name of the action class that is contained in addon

        Attributes:
            _guid (str): a unique GUID used to identify the addon
            _classname (str): the name of the action class that is contained in the addon
            _by (By): Locator strategy for the element associated with this action
            _by_value (str): Corresponding locator strategy value
            _parameters (dict): parameters and their values associated with the action
    """

    def __init__(
        self, guid: str, classname: str,
    ):
        self._guid = guid
        self._classname = classname
        self._by = None
        self._by_value = None
        self._parameters: dict = {}

    @property
    def guid(self) -> str:
        """Getter for the action GUID"""
        return self._guid

    @guid.setter
    def guid(self, value: str):
        """Setter for the action GUID"""
        self._guid = value

    @property
    def classname(self) -> str:
        """Getter for the action class name"""
        return self._classname

    @classname.setter
    def classname(self, value: str):
        """Setter for the action class name"""
        self._classname = value

    @property
    def by(self) -> By:
        """Getter for the element locator strategy"""
        return self._by

    @by.setter
    def by(self, value: By):
        """Setter for the element locator strategy"""
        self._by = value

    @property
    def by_value(self) -> str:
        """Getter for the element locator value"""
        return self._by_value

    @by_value.setter
    def by_value(self, value: str):
        """Setter for the element locator value"""
        self._by_value = value

    @property
    def parameters(self) -> dict:
        """Getter for the action parameters"""
        return self._parameters

    @parameters.setter
    def parameters(self, value: dict):
        """Setter for the action parameters"""
        self._parameters = value



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import inspect

from src.testproject.enums import EnvironmentVariable, ReportNamingElement


class ReportHelper:
    """Provides helper functions used in reporting command, tests and steps"""

    @classmethod
    def infer_test_name(cls) -> str:
        """Tries to infer the test name from the information in the decorator or given to us by pytest or unittest

        Returns:
            str: The inferred test name (typically the test method name)
        """
        # Did we set the test name using our decorator?
        test_name_in_decorator = os.environ.get(EnvironmentVariable.TP_TEST_NAME.value)
        if test_name_in_decorator is not None:
            return test_name_in_decorator

        current_test_info = os.environ.get("PYTEST_CURRENT_TEST")

        if current_test_info is not None:
            # we're using pytest
            result = cls.infer_name_from_pytest_info_for(
                current_test_info, ReportNamingElement.Test
            )
        else:
            # Try finding the right entry in the call stack (for unittest or when no testing framework is used)
            logging.debug("Attempting to infer test name using inspect.stack()")
            result = cls.__find_name_in_call_stack_for(ReportNamingElement.Test)
            logging.debug(f"Inferred test name '{result}' from inspect.stack()")

        return result if result is not None else "Unnamed Test"

    @classmethod
    def infer_project_name(cls) -> str:
        """Tries to infer the project name from the information in the decorator or given to us by pytest or unittest

        Returns:
            str: The inferred project name (typically the folder containing the test file)
        """
        # Did we set the project name using our decorator?
        project_name_in_decorator = os.environ.get(
            EnvironmentVariable.TP_PROJECT_NAME.value
        )
        if project_name_in_decorator is not None:
            return project_name_in_decorator

        current_test_info = os.environ.get("PYTEST_CURRENT_TEST")

        if current_test_info is not None:
            # we're using pytest
            result = cls.infer_name_from_pytest_info_for(
                current_test_info, ReportNamingElement.Project
            )
        else:
            # Try finding the right entry in the call stack (for unittest or when no testing framework is used)
            logging.debug("Attempting to infer project name using inspect.stack()")
            result = cls.__find_name_in_call_stack_for(ReportNamingElement.Project)
            logging.debug(f"Inferred project name '{result}' from inspect.stack()")

        return result if result is not None else "Unnamed Project"

    @classmethod
    def infer_job_name(cls) -> str:
        """Tries to infer the job name from the information in the decorator or given to us by pytest or unittest

        Returns:
            str: The inferred job name (typically the test file name (without the .py extension)
        """
        # Did we set the job name using our decorator?
        job_name_in_decorator = os.environ.get(EnvironmentVariable.TP_JOB_NAME.value)
        if job_name_in_decorator is not None:
            return job_name_in_decorator

        current_test_info = os.environ.get("PYTEST_CURRENT_TEST")

        if current_test_info is not None:
            # we're using pytest
            result = cls.infer_name_from_pytest_info_for(
                current_test_info, ReportNamingElement.Job
            )
        else:
            # Try finding the right entry in the call stack (for unittest or when no testing framework is used)
            logging.debug("Attempting to infer job name using inspect.stack()")
            result = cls.__find_name_in_call_stack_for(ReportNamingElement.Job)
            logging.debug(f"Inferred job name '{result}' from inspect.stack()")

        return result if result is not None else "Unnamed Job"

    @classmethod
    def infer_name_from_pytest_info_for(
        cls, pytest_info: str, element_to_find: ReportNamingElement
    ):
        """Uses the test info stored by pytest to infer a project, job or test name

        Args:
            pytest_info (str): the test info as stored by pytest
            element_to_find (ReportNamingElement): the report naming element that we're looking for

        Returns:
            str: the inferred report naming element value
        """
        if element_to_find == ReportNamingElement.Project:
            path_to_test_file = pytest_info.split(" ")[0].split("::")[0]
            return path_to_test_file[0 : path_to_test_file.rfind("/")].replace("/", ".")
        elif element_to_find == ReportNamingElement.Job:
            path_to_test_file = pytest_info.split(" ")[0].split("::")[0]
            return path_to_test_file.split("::")[0].split("/")[-1].split(".py")[0]
        elif element_to_find == ReportNamingElement.Test:
            return pytest_info.rsplit(" ", maxsplit=1)[0].split("::")[1]
        return None

    @classmethod
    def __find_name_in_call_stack_for(cls, element_to_find: ReportNamingElement) -> str:
        """Uses the current call stack to try and infer a project, job or test name

        Args:
            element_to_find (ReportNamingElement): the report naming element that we're looking for

        Returns:
            str: the inferred report naming element value
        """
        is_unittest = cls.__detect_unittest()

        if is_unittest:
            if element_to_find in [
                ReportNamingElement.Project,
                ReportNamingElement.Job,
            ]:
                # A driver can be initialized inside a test method, but also in a fixture method
                # Therefore we want to look for all these methods when we try to infer project and job names
                # (since project and job names are sent to the Agent upon driver creation)
                for frame in inspect.stack().__reversed__():
                    if frame.function.startswith("test") or frame.function in [
                        "setUp",
                        "tearDown",
                        "setUpClass",
                        "tearDownClass",
                    ]:
                        if element_to_find == ReportNamingElement.Project:
                            path_elements = os.path.normpath(frame.filename).split(
                                os.sep
                            )
                            # return the folder name containing the current test file as the project name
                            return str(path_elements[-2])
                        elif element_to_find == ReportNamingElement.Job:
                            path_elements = os.path.normpath(frame.filename).split(
                                os.sep
                            )
                            # return the current test file name minus the .py extension as the job name
                            return str(path_elements[-1]).split(".py")[0]
                        else:
                            return None
            else:
                # When inferring test names, we are only interested in those methods whose name
                # actually starts with 'test', not in fixture methods
                for frame in inspect.stack().__reversed__():
                    if frame.function.startswith("test"):
                        if element_to_find == ReportNamingElement.Test:
                            # return the current method name as the test name
                            return frame.function
                return None

        else:
            # we're using neither pytest nor unittest, so return sensible values
            inside_module = False

            for frame in inspect.stack().__reversed__():
                if inside_module:
                    if element_to_find == ReportNamingElement.Test:
                        return frame.function
                    elif element_to_find == ReportNamingElement.Project:
                        # in this case we can't infer a project name because there's no data
                        return None
                    elif element_to_find == ReportNamingElement.Job:
                        path_elements = os.path.normpath(frame.filename).split(os.sep)
                        # return the current test file name minus the .py extension as the job name
                        return str(path_elements[-1]).split(".py")[0]
                if str(frame.function) == "<module>":
                    # we're entering the module, the next frame contains the info we're looking for
                    inside_module = True

    @classmethod
    def __detect_unittest(cls) -> bool:
        """Utility method that traverses the call stack and checks if unittest was invoked

        Returns:
            bool: True if unittest was found in the call stack, False otherwise
        """
        for frame in inspect.stack().__reversed__():
            if (
                frame.function == "__init__"
                and str(frame.filename).find("unittest") > 0
                and str(frame.filename).find("main.py") > 0
            ):
                return True
        return False

    @classmethod
    def find_unittest_teardown(cls) -> bool:
        """Utility method that traverses the call stack and checks if unittest was invoked

        Returns:
            bool: True if unittest was found in the call stack, False otherwise
        """
        if not cls.__detect_unittest():
            return False
        else:
            for frame in inspect.stack().__reversed__():
                if frame.function in ["tearDown", "tearDownClass"]:
                    return True
        return False



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class CustomTestReport:
    """Payload object sent to the Agent when reporting a test.

    Args:
        name (str): The test name
        passed (bool): True if the test should be marked as passed, False otherwise
        message (str): A message that goes with the test

    Attributes:
        _name (str): The test name
        _passed (bool): True if the test should be marked as passed, False otherwise
        _message (str): A message that goes with the test
    """

    def __init__(self, name: str, passed: bool, message: str = None):
        self._name = name
        self._passed = passed
        self._message = message

    @property
    def name(self):
        """Getter for the name property"""
        return self._name

    @name.setter
    def name(self, value):
        """Setter for the name property"""
        self._name = value

    @property
    def passed(self):
        """Getter for the passed property"""
        return self._passed

    @passed.setter
    def passed(self, value):
        """Setter for the passed property"""
        self._passed = value

    @property
    def message(self):
        """Getter for the message property"""
        return self._message

    @message.setter
    def message(self, value):
        """Setter for the message property"""
        self._message = value

    def to_json(self):
        """Generates a dict containing the JSON representation of the test payload"""
        return {"name": self._name, "passed": self._passed, "message": self._message}



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from selenium.webdriver.remote.command import Command


class RedactHelper:
    """Class providing helper methods for command redaction

    Args:
        command_executor: The command executor used to send WebDriver commands (Selenium or Appium)

    Attributes:
        _command_executor: The command executor used to send WebDriver commands (Selenium or Appium)
    """

    def __init__(self, command_executor):
        self._command_executor = command_executor

    def redact_command(self, command: str, params: dict):
        """Redacts sensitive contents (passwords) so they do not appear in the reports

        Args:
            command (str): A string specifying the command to execute
            params (dict): A dictionary of named parameters to send with the command as its JSON payload

        Returns:
            dict: A redacted version of the dictionary, where password values are replaced by '****'
        """
        if (
                command == Command.SEND_KEYS_TO_ELEMENT
                or command == Command.SEND_KEYS_TO_ACTIVE_ELEMENT
        ):
            element_id = params["id"]

            if not self._redaction_required(element_id):
                return params

            # Change text typed into redactable field to '***'
            params["text"] = "***"
            params["value"] = list("***")

        return params

    def _redaction_required(self, element_id: str) -> bool:
        """Checks if the element should be redacted

        Args:
            element_id (str): The ID of the element under investigation

        Returns:
            bool: True if the element should be redacted, False otherwise
        """
        capabilities = self._command_executor.agent_client.agent_session.capabilities
        platform_name = capabilities.get("platformName")
        browser_name = capabilities.get("browserName")

        # Check if element is a mobile password element
        if platform_name.casefold() == "android":
            # Check that we're not dealing with mobile web
            if browser_name is None or browser_name == "":
                return self._is_android_password_element(element_id)

        return self._is_secured_element(element_id)

    def _is_android_password_element(self, element_id: str) -> bool:
        """Checks if the element is an Android password element

        Args:
            element_id (str): The ID of the element under investigation

        Returns:
            bool: True if the element is an Android password element, False otherwise
        """
        get_attribute_params = {
            "sessionId": self._command_executor.agent_client.agent_session.session_id,
            "id": element_id,
            "name": "password",
        }
        get_attribute_response = self._command_executor.execute(
            Command.GET_ELEMENT_ATTRIBUTE, get_attribute_params, True
        )

        return get_attribute_response.get("value").casefold() == "true"

    def _is_secured_element(self, element_id: str) -> bool:
        """Checks if the element is a secured element (an HTML or iOS password element)

        Args:
            element_id (str): The ID of the element under investigation

        Returns:
            bool: True if the element is a secured element, False otherwise
        """
        get_attribute_params = {
            "sessionId": self._command_executor.agent_client.agent_session.session_id,
            "id": element_id,
            "name": "type",
        }
        get_attribute_response = self._command_executor.execute(
            Command.GET_ELEMENT_ATTRIBUTE, get_attribute_params, True
        )
        return get_attribute_response["value"] in ["password", "XCUIElementTypeSecureTextField"]



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from src.testproject.sdk.exceptions.notimplementedexception import (
    NotImplementedException,
)
from src.testproject.sdk.internal.agent import AgentClient
from src.testproject.sdk.internal.helpers.reporting_command_executor import (
    ReportingCommandExecutor,
)


class GenericCommandExecutor(ReportingCommandExecutor):
    """Custom command executor class for the generic driver

    Args:
        agent_client (AgentClient): Client used to communicate with the TestProject Agent
    """

    def __init__(self, agent_client: AgentClient):
        ReportingCommandExecutor.__init__(
            self, agent_client=agent_client, command_executor=self
        )

    def execute(self, command: str, params: dict, skip_reporting: bool = False):
        """Not implemented for generic driver objects

        Args:
            command (str): A string specifying the command to execute
            params (dict): A dictionary of named parameters to send with the command as its JSON payload
            skip_reporting (bool): True if command should not be reported to Agent, False otherwise
        """
        raise NotImplementedException("Not implemented for generic driver")



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ResultField:
    """Object representing a field associated with an action

        Attributes:
            _name (str): The field name
            _value: the field value
            _is_output (bool): True if the field is an output field, False otherwise
    """

    def __init__(self):
        self._name = None
        self._value = None
        self._is_output = None

    @property
    def name(self) -> str:
        """Getter for the field name"""
        return self._name

    @name.setter
    def name(self, value: str):
        """Setter for the field name"""
        self._name = value

    @property
    def value(self):
        """Getter for the field value"""
        return self._value

    @value.setter
    def value(self, val):
        """Setter for the field value"""
        self._value = val

    @property
    def is_output(self) -> bool:
        """Getter for the field is_output indication"""
        return self._is_output

    @is_output.setter
    def is_output(self, value: bool):
        """Setter for the field is_output indication"""
        self._is_output = value



from .chrome import Chrome
from .firefox import Firefox
from .edge import Edge
from .safari import Safari
from .ie import Ie
from .remote import Remote
from .generic import Generic

__all__ = ["Chrome", "Firefox", "Edge", "Safari", "Ie", "Remote", "Generic"]



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Class containing all the GUIDs uniquely defining the testproject() actions"""

actions = {"PAUSE_ID": "5ac5d621-7a53-45a6-94e5-a3348be3969e"}

driver_actions = {
    "SEND_KEYS_ID": "928bf14e-7a56-4371-b92a-9cb9de53f632",
    "IS_SELECTED_ID": "2c399e26-8d02-41a7-a8d4-acfeda14fb5b",
    "IS_PRESENT_ID": "a9ab4ae7-38ca-4451-a636-1b801a774503",
    "IS_VISIBLE_ID": "7314f516-588b-4dff-bf0e-7016c7e02a4b",
    "CONTAINS_TEXT_ID": "85da72d0-c5a4-4587-be33-8cb3b5fa7910",
    "IS_CLICKABLE_ID": "a59f3a20-d08d-4293-9add-b279f92e6c41",
    "CLICK_IF_VISIBLE_ID": "27fd3f33-9b63-4dc4-97e9-fcf494cbffdb",
    "IS_INVISIBLE_ID": "104373b7-868c-44ba-bbd2-e16089ad2e48",
    "TYPE_TEXT_ID": "6f94fdaa-0041-4265-b73b-ecfcbe702d66",
    "CLEAR_CONTENTS_ID": "511ddb4e-210f-4f3f-9fb9-b766bb23ee6a",
    "GET_TEXT_ID": "225caec3-3231-45d5-a3f3-c4c59b7ab385",
    "CLICK_ID": "02d572b2-d610-4700-9ad3-40076ddb9807",
    "GET_TITLE_ID": "c168339f-55c4-4c0d-91fc-cd9d4b8c7574",
}

web_actions = {
    "MOVE_MOUSE_TO_ELEMENT_ID": "631118c0-4828-42c8-aa69-0e6107d5956d",
    "NAVIGATE_FORWARD_ID": "664577f4-e1d2-4c46-b499-3934b447b478",
    "NAVIGATE_BACK_ID": "43b4408f-3fa1-4ec5-80cd-872d6dff7628",
    "REFRESH_ID": "3a06acbf-9138-4392-bca1-c1946886b65d",
    "NAVIGATE_TO_URL_ID": "49c5d8d8-4ba8-42f0-823d-702e6acfb23a",
    "GET_CURRENT_URL_ID": "56a4ad1b-b563-4322-ba29-d165090302b2",
    "SCROLL_WINDOW_ID": "a5d4eb31-51f0-4c62-8832-60c6bcd51a16",
    "SELECT_ALL_OPTIONS_BY_VALUE_ID": "45fdd921-b359-4d6a-b7c5-5f9a3612c0f2",
    "SWITCH_TO_WINDOW_ID": "3a7a12fb-66a4-47c2-aaf0-deeb7b1f748a",
    "CLOSE_WINDOW_ID": "6cc71ddb-1aaf-433c-aa94-56d944aac433",
}



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os

from src.testproject.enums import EnvironmentVariable


def report(project: str = None, job: str = None, test: str = None):
    """Enables setting a custom name for the project, job and test for reporting purposes
        Args:
            project (str): The name of the project
            job (str): The name of the job
            test (str): The name of the test
        Returns:
            report_decorator: The decorated test method.
    """

    def report_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if project is not None:
                os.environ[EnvironmentVariable.TP_PROJECT_NAME.value] = project
            if job is not None:
                os.environ[EnvironmentVariable.TP_JOB_NAME.value] = job
            if test is not None:
                os.environ[EnvironmentVariable.TP_TEST_NAME.value] = test
            return func(*args, **kwargs)

        return wrapper

    return report_decorator



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from src.testproject.helpers import ConfigHelper
from src.testproject.rest import ReportSettings


class SessionRequest:
    """Represent a request message object to be sent to the Agent to initialize a new session

    Args:
        capabilities (dict): Desired session capabilities
        reportsettings (ReportSettings): Settings to be used in the report

    Attributes:
        _capabilities (dict): Desired session capabilities
        _sdk_version (str): Current Python SDK version
        _language (str): Test code language (Python, obviously)
        _projectname (str): Project name to report
        _jobname (str): Job name to report
    """

    def __init__(self, capabilities: dict, reportsettings: ReportSettings):
        self._capabilities: dict = capabilities
        self._sdk_version = ConfigHelper.get_sdk_version()
        self._language = "Python"
        self._projectname = reportsettings.projectname
        self._jobname = reportsettings.jobname

    def to_json(self):
        """Returns a JSON representation of the current SessionRequest instance"""
        return {
            "projectName": self._projectname,
            "jobName": self._jobname,
            "capabilities": self._capabilities,
            "sdkVersion": self._sdk_version,
            "language": self._language,
        }



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from selenium.webdriver.common.by import By
from src.testproject.classes import ElementSearchCriteria
from src.testproject.enums import FindByType
from src.testproject.sdk.exceptions import SdkException


class SeleniumHelper:
    """Contains helper methods for Selenium actions, mostly locator-related"""

    @staticmethod
    def create_search_criteria(by: By, by_value: str):
        """Translator method to create element search criteria to send to the Agent

        Args:
            by (By): The element locator strategy to be used
            by_value (str): The associated element locator strategy value

        Returns:
            ElementSearchCriteria: object representing the element search criteria
        """

        if by == By.ID:
            return ElementSearchCriteria(FindByType.ID, by_value)
        elif by == By.NAME:
            return ElementSearchCriteria(FindByType.NAME, by_value)
        elif by == By.XPATH:
            return ElementSearchCriteria(FindByType.XPATH, by_value)
        elif by == By.CLASS_NAME:
            return ElementSearchCriteria(FindByType.CLASSNAME, by_value)
        elif by == By.CSS_SELECTOR:
            return ElementSearchCriteria(FindByType.CSSSELECTOR, by_value)
        elif by == By.LINK_TEXT:
            return ElementSearchCriteria(FindByType.LINKTEXT, by_value)
        elif by == By.PARTIAL_LINK_TEXT:
            return ElementSearchCriteria(FindByType.PARTIALLINKTEXT, by_value)
        elif by == By.TAG_NAME:
            return ElementSearchCriteria(FindByType.TAG_NAME, by_value)
        else:
            raise SdkException(f"Did not recognize locator strategy {by}")

    @staticmethod
    def create_addon_locator(by: By, by_value: str):
        """Creates and returns an locator used in an addon based on a locator strategy

        Args:
            by (By): The element locator strategy to be used
            by_value (str): The associated element locator strategy value

        Returns:
            dict: object representing the element locator strategy to use in the addon
        """

        if by == By.ID:
            return {"id": by_value}
        elif by == By.NAME:
            return {"name": by_value}
        elif by == By.XPATH:
            return {"xpath", by_value}
        elif by == By.CLASS_NAME:
            return {"className": by_value}
        elif by == By.CSS_SELECTOR:
            return {"cssSelector": by_value}
        elif by == By.LINK_TEXT:
            return {"linkText": by_value}
        elif by == By.PARTIAL_LINK_TEXT:
            return {"partialLinkText": by_value}
        elif by == By.TAG_NAME:
            return {"tagName": by_value}
        else:
            raise SdkException(f"Did not recognize locator strategy {by}")



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from src.testproject.enums import ExecutionResultType
from src.testproject.sdk.drivers.actions import Actions
from src.testproject.sdk.internal.agent import AgentClient
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from src.testproject.sdk.drivers.actions.action_guids import driver_actions


class DriverActions(Actions):
    """Offers methods to execute driver actions

    Args:
        agent_client (AgentClient): client to communicate with the Agent
        timeout (int): timeout for action execution
    """

    def __init__(self, agent_client: AgentClient, timeout: int):
        super().__init__(agent_client, timeout)

    def send_keys_to_window(self, keys) -> bool:
        """Send a list of keystrokes to the current browser

        Args:
            keys: list of keystrokes

        Returns:
            bool: True if the action was performed successfully, False otherwise
        """
        typable_keys = self.__convert_to_typable(keys)
        body = {"text": ",".join(typable_keys)}
        response = self.action_execute(driver_actions["SEND_KEYS_ID"], body, None, "")
        return response.executionresulttype == ExecutionResultType.Passed

    def is_selected(self, by: By, by_value: str) -> bool:
        """Checks if an element is selected

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: True if the element is selected, False otherwise
        """
        response = self.action_execute(driver_actions["IS_SELECTED_ID"], {}, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def is_present(self, by: By, by_value: str) -> bool:
        """Checks if an element is present in the DOM

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: True if the element is present in the DOM, False otherwise
        """
        response = self.action_execute(driver_actions["IS_PRESENT_ID"], {}, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def is_visible(self, by: By, by_value: str) -> bool:
        """Checks if an element is visible

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: True if the element is visible, False otherwise
        """
        response = self.action_execute(driver_actions["IS_VISIBLE_ID"], {}, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def contains_text(self, text_to_find: str, by: By, by_value: str) -> bool:
        """Checks if the text of an element contains a given substring

        Args:
            text_to_find (str): The substring to find in the element text
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: True if the element text contains the given substring, False otherwise
        """
        body = {"text": text_to_find}
        response = self.action_execute(driver_actions["CONTAINS_TEXT_ID"], body, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def is_clickable(self, by: By, by_value: str) -> bool:
        """Checks if an element is clickable

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: True if the element is clickable, False otherwise
        """
        response = self.action_execute(driver_actions["IS_CLICKABLE_ID"], {}, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def is_invisible(self, by: By, by_value: str) -> bool:
        """Checks if an element is invisible (or not present in the DOM)

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: True if the element is invisible, False otherwise
        """
        response = self.action_execute(driver_actions["IS_INVISIBLE_ID"], {}, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def send_keys(self, text_to_type: str, by: By, by_value: str) -> bool:
        """Sends the given text to the specified element

        Args:
            text_to_type (str): The text to type in the element
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: Element text if the action was successful, None otherwise
        """
        body = {"keys": text_to_type}
        response = self.action_execute(driver_actions["TYPE_TEXT_ID"], body, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def clear_contents(self, by: By, by_value: str) -> bool:
        """Clears the contents of an element

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: Element text if the element was found, None otherwise
        """
        response = self.action_execute(driver_actions["CLEAR_CONTENTS_ID"], {}, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def get_text(self, by: By, by_value: str) -> str:
        """Retrieves the visible text of an element

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: Element text if the element was found, None otherwise
        """
        response = self.action_execute(driver_actions["GET_TEXT_ID"], {}, by, by_value)
        if response.executionresulttype != ExecutionResultType.Passed:
            return None
        return response.outputs["text"]

    def click(self, by: By, by_value: str) -> bool:
        """Clicks an element

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        response = self.action_execute(driver_actions["CLICK_ID"], {}, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def get_title(self) -> str:
        """Retrieves the current driver or application title

        Returns:
            bool: Driver or application title if successful, None otherwise
        """
        response = self.action_execute(driver_actions["GET_TITLE_ID"], {})
        if response.executionresulttype != ExecutionResultType.Passed:
            return None
        return response.outputs["title"]

    @staticmethod
    def __convert_to_typable(keys):
        """Converts a list of characters to a typable representation

        Returns:
            list: typable version of the list of characters
        """
        typing = []
        for key in keys:
            if isinstance(key, Keys):
                typing.append(key)
            elif isinstance(key, int):
                key = str(key)
                for i in range(len(key)):
                    typing.append(key[i])
            else:
                for i in range(len(key)):
                    typing.append(key[i])
        return typing



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from src.testproject import definitions
from src.testproject.sdk.exceptions import SdkException


class ConfigHelper:
    """Contains helper methods for SDK configuration"""

    @staticmethod
    def get_agent_service_address() -> str:
        """Returns the Agent service address as defined in the TP_AGENT_URL environment variable.
            Defaults to http://127.0.0.1:8585 (localhost)

        Returns:
            str: the Agent service address
        """
        address = os.getenv("TP_AGENT_URL")
        if address is None:
            logging.info(
                "No Agent service address found in TP_AGENT_URL environment variable, "
                "defaulting to http://127.0.0.1:8585 (localhost)"
            )
            return "http://127.0.0.1:8585"
        # Replace 'localhost' with '127.0.0.1' to prevent delays as a result of DNS lookups
        address = address.replace("localhost", "127.0.0.1")
        logging.info(f"Using {address} as the Agent URL")
        return address

    @staticmethod
    def get_developer_token() -> str:
        """Returns the TestProject developer token as defined in the TP_DEV_TOKEN environment variable

        Returns:
            str: the developer token
        """
        token = os.getenv("TP_DEV_TOKEN")
        if token is None:
            logging.error("No developer token was found, did you set it in the TP_DEV_TOKEN environment variable?")
            logging.error("You can get a developer token from https://app.testproject.io/#/integrations/sdk?lang=Python")
            raise SdkException("No development token defined in TP_DEV_TOKEN environment variable")
        return token

    @staticmethod
    def get_sdk_version() -> str:
        """Returns the SDK version as defined in the definitions module

        Returns:
            str: the current SDK version
        """
        return definitions.get_sdk_version()



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, unique


@unique
class FindByType(Enum):
    """Enumeration of supported element locator strategies"""
    ID = 0
    NAME = 1
    CLASSNAME = 2
    CSSSELECTOR = 3
    LINKTEXT = 4
    PARTIALLINKTEXT = 5
    TAG_NAME = 6
    XPATH = 7
    ACCESSIBILITYID = 8
    IOSUIAUTOMATION = 9
    ANDROIDUIAUTOMATOR = 10
    NAMEDTEXTFIELD = 11



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class OperationResult:
    """Represents the result of an operation performed by the Agent

    Args:
        passed (bool): True if the operation was executed successfully, False otherwise
        status_code (int): The HTTP status code of the Agent response
        message (str): A message returned by the Agent after the operation execution
        data (str): Output data for the operation performed

    Attributes:
        _passed (bool): True if the operation was executed successfully, False otherwise
        _status_code (int): The HTTP status code of the Agent response
        _message (str): A message returned by the Agent after the operation execution
        _data (str): Output data for the operation performed

    """
    def __init__(self, passed: bool = False, status_code: int = 500, message: str = "", data=None):
        self._passed = passed
        self._status_code = status_code
        self._message = message
        self._data = data

    @property
    def passed(self):
        """Getter for the pass / fail indication"""
        return self._passed

    @passed.setter
    def passed(self, value):
        """Setter for the pass / fail indication"""
        self._passed = value

    @property
    def status_code(self):
        """Getter for the response status code"""
        return self._status_code

    @status_code.setter
    def status_code(self, value):
        """Setter for the response status code"""
        self._status_code = value

    @property
    def message(self):
        """Getter for the operation response message"""
        return self._message

    @message.setter
    def message(self, value):
        """Setter for the operation response message"""
        self._message = value

    @property
    def data(self):
        """Getter for the operation output data"""
        return self._data

    @data.setter
    def data(self, value):
        """Setter for the operation output data"""
        self._data = value



from .sessionrequest import SessionRequest
from .sessionresponse import SessionResponse
from .drivercommandreport import DriverCommandReport
from .stepreport import StepReport
from .customtestreport import CustomTestReport
from .addonexecutionresponse import AddonExecutionResponse

__all__ = [
    "SessionRequest",
    "SessionResponse",
    "DriverCommandReport",
    "StepReport",
    "CustomTestReport",
    "AddonExecutionResponse",
]



from .seleniumhelper import SeleniumHelper
from .confighelper import ConfigHelper
from .sockethelper import SocketHelper
from .reporthelper import ReportHelper
from .logginghelper import LoggingHelper

__all__ = [
    "SeleniumHelper",
    "ConfigHelper",
    "SocketHelper",
    "ReportHelper",
    "LoggingHelper",
]



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import inspect

from src.testproject.classes import ActionExecutionResponse
from src.testproject.enums import ExecutionResultType
from src.testproject.helpers import SeleniumHelper
from src.testproject.sdk.internal.agent import AgentClient
from selenium.webdriver.common.by import By
from src.testproject.sdk.drivers.actions.action_guids import actions


class Actions:
    """Offers method to execute all actions

    Args:
        agent_client (AgentClient): client to communicate with the Agent
        timeout (int): timeout for action execution

    Attributes:
        _agent_client (AgentClient): client to communicate with the Agent
        _timeout (int): timeout for action execution
    """
    def __init__(self, agent_client: AgentClient, timeout: int):
        self._agent_client = agent_client
        self._timeout = timeout

    def action_execute(
        self, action_guid: str, body: dict, by: By = None, by_value: str = "", timeout: int = 10000,
    ) -> ActionExecutionResponse:
        """Sends HTTP request to Agent

            Args:
                action_guid (str): The TestProject action GUID to be executed
                body (dict): Parameters to be passed to the Agent
                by (By): The locator strategy to be used to locate the element to perform the action on/with
                by_value (str): The associated locator strategy value
                timeout (int): timeout (in seconds) for the action execution

            Returns:
                ActionExecutionResponse: contains result of the sent execution request
        """
        body["_timeout"] = timeout

        if by is not None:
            search_criteria = SeleniumHelper.create_search_criteria(by, by_value)
            if search_criteria is not None:
                body["elementSearchCriteria"] = search_criteria.to_json()
            else:
                logging.error(f"Failure in creating search criteria from locator strategy {by} with value {by_value}")

        response = self._agent_client.send_action_execution_request(action_guid, body)
        if response.executionresulttype == ExecutionResultType.Failed:
            logging.warning(
                f"Failed to execute action '{inspect.stack()[1].function}', "
                "agent returned the following message: {response.message}"
            )
        return response

    def pause(self, milliseconds: int) -> bool:
        """Pause test execution for the specified duration

        Args:
            milliseconds (int): Number of milliseconds to pause test execution for

        Returns:
            bool: True if the action was performed successfully, False otherwise
        """
        body = {"milliseconds": milliseconds}
        response = self.action_execute(actions["PAUSE_ID"], body)
        return response.executionresulttype == ExecutionResultType.Passed



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, unique


@unique
class ExecutionResultType(Enum):
    """Enumeration of possible action execution results"""
    NoResult = 0
    Executing = 1
    Skipped = 2
    Passed = 3
    Failed = 4
    Disabled = 5
    Aborted = 6
    Suspended = 7
    Error = 8



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid


class StepReport:
    """Payload object sent to the Agent when reporting a test step.

        Args:
            description (str): The step description
            message (str): A message that goes with the step
            passed (bool): True if the step should be marked as passed, False otherwise
            screenshot (str): A base64 encoded screenshot that is associated with the step

        Attributes:
            _description (str): The step description
            _message (str): A message that goes with the step
            _passed (bool): True if the step should be marked as passed, False otherwise
            _screenshot (str): A base64 encoded screenshot that is associated with the step
    """

    def __init__(self, description: str, message: str, passed: bool, screenshot: str = None):
        self._description = description
        self._message = message
        self._passed = passed
        self._screenshot = screenshot

    def to_json(self) -> dict:
        """Generates a dict containing the JSON representation of the step payload"""
        json = {
            "guid": str(uuid.uuid4()),
            "description": self._description,
            "message": self._message,
            "passed": self._passed,
        }
        if self._screenshot is not None:
            json["screenshot"] = self._screenshot

        return json



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from selenium.webdriver import DesiredCapabilities

from src.testproject.sdk.drivers.webdriver.base import BaseDriver


class Safari(BaseDriver):
    """Used to create a new Safari browser instance

    Args:
        desired_capabilities (dict): Safari automation session desired capabilities and options
        token (str): The developer token used to communicate with the agent
        projectname (str): Project name to report
        jobname (str): Job name to report
        disable_reports (bool): set to True to disable all reporting (no report will be created on TestProject)
    """

    def __init__(
        self,
        desired_capabilities: dict = DesiredCapabilities.SAFARI,
        token: str = None,
        projectname: str = None,
        jobname: str = None,
        disable_reports: bool = False,
    ):
        super().__init__(
            capabilities=desired_capabilities,
            token=token,
            projectname=projectname,
            jobname=jobname,
            disable_reports=disable_reports,
        )



from .executionresulttype import ExecutionResultType
from .executionfailuretype import ExecutionFailureType
from .findbytype import FindByType
from .reportnamingelement import ReportNamingElement
from .environmentvariable import EnvironmentVariable

__all__ = [
    "ExecutionResultType",
    "ExecutionFailureType",
    "FindByType",
    "ReportNamingElement",
    "EnvironmentVariable",
]



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AgentSession:
    """Object representing the current Agent session

    Args:
        remote_address (str): The Agent address
        session_id (str): A unique identifier for the current Agent session
        dialect (str): The WebDriver dialect associated with the session (W3C or OSS)
        capabilities (str): driver-specific capabilities

    Attributes:
        _remote_address (str): The Agent address
        _session_id (str): A unique identifier for the current Agent session
        _dialect (str): The WebDriver dialect associated with the session (W3C or OSS)
        _capabilities (str): driver-specific capabilities
    """

    def __init__(self, remote_address: str, session_id: str, dialect: str, capabilities: dict):
        self._remote_address = remote_address
        self._session_id = session_id
        self._dialect = dialect
        self._capabilities = capabilities

    @property
    def remote_address(self) -> str:
        """Getter for the Agent address"""
        return self._remote_address

    @property
    def session_id(self) -> str:
        """Getter for the current session ID"""
        return self._session_id

    @property
    def dialect(self) -> str:
        """Getter for the session dialect"""
        return self._dialect

    @property
    def capabilities(self) -> dict:
        """Getter for the driver capabilities"""
        return self._capabilities



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from src.testproject.enums import ExecutionResultType


class ActionExecutionResponse:
    """Defines an object type for responses returned by the Agent upon action execution

    Args:
        executionresulttype (ExecutionResultType): The execution result
        message (str): A message returned by the Agent after action execution
        outputs (dict): Results of the action execution in key-value pairs

    Attributes:
        _executionresulttype (ExecutionResultType): The execution result
        _message (str): A message returned by the Agent after action execution
        _outputs (dict): Results of the action execution in key-value pairs
    """
    def __init__(
        self, executionresulttype: ExecutionResultType = ExecutionResultType.NoResult, message: str = "", outputs: dict = None,
    ):
        self._executionresulttype = executionresulttype
        self._message = message
        self._outputs = outputs

    @property
    def executionresulttype(self) -> ExecutionResultType:
        """Getter for the action execution result type"""
        return self._executionresulttype

    @executionresulttype.setter
    def executionresulttype(self, value: ExecutionResultType):
        """Setter for the action execution result type"""
        self._executionresulttype = value

    @property
    def message(self) -> str:
        """Getter for the action execution response message"""
        return self._message

    @message.setter
    def message(self, value: str):
        """Setter for the action execution response message"""
        self._message = value

    @property
    def outputs(self) -> dict:
        """Getter for the action execution response outputs"""
        return self._outputs

    @outputs.setter
    def outputs(self, value: dict):
        """Setter for the action execution response outputs"""
        self._outputs = value



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class AgentStatusResponse:
    """Represent a response message object returned by the Agent when requesting its status

    Args:
        tag (str): The Agent version

    Attributes:
        _tag (str): The Agent version
    """

    def __init__(self, tag: str):
        self._tag = tag

    @property
    def tag(self) -> str:
        """Getter for the Agent version field ('tag' in the JSON response)"""
        return self._tag



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os

from src.testproject.helpers import ReportHelper
from src.testproject.rest.messages import StepReport, CustomTestReport


class Reporter:
    """Exposes reporting actions to the WebDriver object

    Args:
        command_executor: the command executor associated with the driver

    Attributes:
        _command_executor: the command executor associated with the driver
    """

    def __init__(self, command_executor):
        self._command_executor = command_executor

    def step(
        self, description: str, message: str, passed: bool, screenshot: bool = False
    ):
        """Sends a step report to the Agent Client

        Args:
            description (str): The step description
            message (str): A message that goes with the step
            passed (bool): True if the step should be marked as passed, False otherwise
            screenshot (bool): True if a screenshot should be made, False otherwise
        """

        # First update the current test name and report a test if necessary
        self._command_executor.update_known_test_name()

        if not self._command_executor.disable_reports:

            step_report = StepReport(
                description,
                message,
                passed,
                self._command_executor.create_screenshot() if screenshot else None,
            )
            self._command_executor.agent_client.report_step(step_report)
        else:
            logging.debug(f"Step '{description}' {'passed' if passed else 'failed'}")

    def test(self, name: str = None, passed: bool = True, message: str = None):
        """Sends a test report to the Agent Client

        Args:
            name (str): The test name
            passed (bool): True if the test should be marked as passed, False otherwise
            message (str): A message that goes with the test
        """
        if not self._command_executor.disable_reports:
            if name is None:
                name = ReportHelper.infer_test_name()

            if not self._command_executor.disable_auto_test_reports:
                if os.getenv("RFW_SUPPRESS_WARNINGS", "false").casefold() != "true":
                    logging.warning(
                        "Automatic reporting is enabled, disable this using disable_reports flag "
                        "when creating a driver instance to avoid duplicates in the report"
                    )

            test_report = CustomTestReport(name=name, passed=passed, message=message,)

            self._command_executor.agent_client.report_test(test_report)

        else:
            logging.debug(f"Test '{name}' {'passed' if passed else 'failed'}")

    def disable_reports(self, disabled: bool):
        """Enables or disabled all reporting

        Args:
            disabled (bool): Set to True to disable all reporting to the Agent
        """
        self._command_executor.disable_reports = disabled

    def disable_auto_test_reports(self, disabled: bool):
        """Enables or disables automatically reporting a test to the Agent

        Args:
            disabled (bool): Set to True to disable automatically reporting a test to the Agent
        """
        self._command_executor.disable_auto_test_reports = disabled

    def disable_command_reports(self, disabled: bool):
        """Enables or disables driver command reporting

        Args:
            disabled (bool): Set to True to disable driver command reports being sent to the Agent.
        """
        self._command_executor.disable_command_reports = disabled

    def disable_redaction(self, disabled: bool):
        """Enables or disables driver command report redaction

        Args:
            disabled (bool): Set to True to disable driver command report redaction.
        """
        self._command_executor.disable_redaction = disabled

    def exclude_test_names(self, excluded_test_names: list):
        """Excludes a list of test names (as strings) from being reported

        Args:
            excluded_test_names: list of strings containing tests that should not be reported.
        """
        self._command_executor.excluded_test_names = excluded_test_names



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from src.testproject.enums import EnvironmentVariable
from src.testproject.helpers import ReportHelper, LoggingHelper, ConfigHelper
from src.testproject.rest import ReportSettings
from src.testproject.sdk.exceptions import SdkException
from src.testproject.sdk.internal.agent import AgentClient
from src.testproject.sdk.internal.helpers import CustomCommandExecutor
from src.testproject.sdk.internal.reporter import Reporter
from src.testproject.sdk.internal.session import AgentSession
from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver


class BaseDriver(RemoteWebDriver):
    """Base class with common functions for all web browser types

    Args:
        capabilities (dict): Automation session desired capabilities and options
        token (str): Developer token to be used to communicate with the Agent
        projectname (str): Project name to report
        jobname (str): Job name to report
        disable_reports (bool): set to True to disable all reporting (no report will be created on TestProject)

    Attributes:
        _agent_client (AgentClient): client responsible for communicating with the TestProject agent
        _agent_session (AgentSession): stores properties of the current agent session
        command_executor (CustomCommandExecutor): the HTTP command executor used to send instructions
        to remote WebDriver
        w3c (bool): indicates whether or not the driver instance uses the W3C dialect
        session_id (str): contains the current session ID
    """

    __instance = None

    def __init__(
        self,
        capabilities: dict,
        token: str,
        projectname: str,
        jobname: str,
        disable_reports: bool,
    ):

        if BaseDriver.__instance is not None:
            raise SdkException("A driver session already exists")

        LoggingHelper.configure_logging()

        if token is not None:
            logging.info(f"Token used as specified in constructor: {token}")

        self._token = token if token is not None else ConfigHelper.get_developer_token()

        if disable_reports:
            # Setting the project and job name to empty strings will cause the Agent to not initialize a report
            self._projectname = ""
            self._jobname = ""
        else:
            self._projectname = (
                projectname
                if projectname is not None
                else ReportHelper.infer_project_name()
            )
            self._jobname = (
                jobname if jobname is not None else ReportHelper.infer_job_name()
            )

        self._agent_client: AgentClient = AgentClient(
            token=self._token,
            capabilities=capabilities,
            reportsettings=ReportSettings(self._projectname, self._jobname),
        )
        self._agent_session: AgentSession = self._agent_client.agent_session
        self.w3c = True if self._agent_session.dialect == "W3C" else False

        # Create a custom command executor to enable:
        # - automatic logging capabilities
        # - customized reporting settings
        self.command_executor = CustomCommandExecutor(
            agent_client=self._agent_client,
            remote_server_addr=self._agent_session.remote_address,
        )

        self.command_executor.disable_reports = disable_reports

        RemoteWebDriver.__init__(
            self,
            command_executor=self.command_executor,
            desired_capabilities=self._agent_session.capabilities,
        )

        BaseDriver.__instance = self

    @classmethod
    def instance(cls):
        """Returns the singleton instance of the driver object"""
        return cls.__instance

    def start_session(self, capabilities, browser_profile=None):
        """Sets capabilities and sessionId obtained from the Agent when creating the original session."""
        self.session_id = self._agent_session.session_id
        logging.info(f"Session ID is {self.session_id}")

    def report(self) -> Reporter:
        """Enables access to the TestProject reporting actions from the driver object

        Returns:
            Reporter: object giving access to reporting methods
        """
        return Reporter(self.command_executor)

    def quit(self):
        """Quits the driver and stops the session with the Agent, cleaning up after itself"""
        # Report any left over driver command reports
        self.command_executor.clear_stash()

        # Make instance available again
        BaseDriver.__instance = None

        try:
            RemoteWebDriver.quit(self)
        except Exception:
            pass

        # Stop the Agent client
        self.command_executor.agent_client.stop()

        # Clean up any environment variables set in the decorator
        for env_var in [
            EnvironmentVariable.TP_TEST_NAME,
            EnvironmentVariable.TP_PROJECT_NAME,
            EnvironmentVariable.TP_JOB_NAME,
        ]:
            EnvironmentVariable.remove(env_var)



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os


class LoggingHelper:
    """Contains helper methods for setting custom logging property values"""

    @staticmethod
    def configure_logging():
        """Configures logging according to values set in dedicated environment variables"""

        logging_level = os.environ.get("TP_LOG_LEVEL", "INFO")
        logging_format = os.environ.get("TP_LOG_FORMAT", "%(asctime)s %(levelname)s %(message)s")

        logging.basicConfig(level=logging_level, format=logging_format)



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, unique


@unique
class ExecutionFailureType(Enum):
    """Enumeration of possible command execution failure reasons"""
    NoFailure = 0
    DriverError = 1
    MalformedExecutionJsonObject = 2
    ConditionsNotMet = 3
    RemoteResourceNotAvailable = 4
    InnerException = 5
    ValidationFailed = 6
    MissingEnvironmentCapabilities = 7
    StepCausedTestFailure = 8
    SecurityIssu = 9



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import socket
import requests
import threading
import queue
import uuid

from urllib.parse import urljoin
from enum import Enum, unique
from requests import HTTPError
from src.testproject.classes import ActionExecutionResponse
from src.testproject.enums import ExecutionResultType
from src.testproject.executionresults import OperationResult
from src.testproject.helpers import ConfigHelper
from src.testproject.rest import ReportSettings
from src.testproject.rest.messages import (
    SessionRequest,
    SessionResponse,
    DriverCommandReport,
    StepReport,
    CustomTestReport,
)
from src.testproject.rest.messages.agentstatusresponse import AgentStatusResponse
from src.testproject.sdk.exceptions import (
    SdkException,
    AgentConnectException,
    InvalidTokenException,
    ObsoleteVersionException,
    MissingBrowserException,
)
from src.testproject.helpers import SocketHelper
from src.testproject.sdk.internal.session import AgentSession


class AgentClient:
    """Client used to communicate with the TestProject Agent process

    Args:
        token (str): The development token used to communicate with the Agent
        capabilities (dict): Additional options to be applied to the driver instance
        reportsettings (ReportSettings): Settings (project name, job name) to be included in the report

    Attributes:
        _remote_address (str): The Agent endpoint
        _capabilities (dict): Additional options to be applied to the driver instance
        _agent_session (AgentSession): stores properties of the current agent session
        _token (str): The development token used to authenticate with the Agent
        _reportsettings (ReportSettings): Settings (project name, job name) to be included in the report
        _sock (socket.Socket): Socket used in communicating with the Agent
        _queue (queue.Queue): queue holding reports to be sent to Agent in separate thread
    """

    REPORTS_QUEUE_TIMEOUT = 10

    def __init__(self, token: str, capabilities: dict, reportsettings: ReportSettings):
        self._remote_address = ConfigHelper.get_agent_service_address()
        self._capabilities = capabilities
        self._agent_session = None
        self._token = token
        self._reportsettings = reportsettings
        self._queue = queue.Queue()

        self._running = True
        self._reporting_thread = threading.Thread(
            target=self.__report_worker, daemon=True
        )
        self._reporting_thread.start()

        if not self.__start_session():
            raise SdkException("Failed to start development mode session")

    @property
    def agent_session(self):
        """Getter for the Agent session object"""
        return self._agent_session

    def __start_session(self) -> bool:
        """Starts a new development session with the Agent

        Returns:
            bool: True if the session started successfully, False otherwise
        """
        sdk_version = ConfigHelper.get_sdk_version()

        logging.info(f"SDK version: {sdk_version}")

        start_session_response = self._request_session_from_agent()

        self._agent_session = AgentSession(
            start_session_response.server_address,
            start_session_response.session_id,
            start_session_response.dialect,
            start_session_response.capabilities,
        )

        self._sock = SocketHelper.create_connection(
            self._remote_address, start_session_response.dev_socket_port
        )

        logging.info("Development session started...")
        return True

    def _request_session_from_agent(self) -> SessionResponse:
        """Creates and sends a session request object

        Returns:
            SessionResponse: object containing the response to the session request
        """
        session_request = SessionRequest(self._capabilities, self._reportsettings)

        logging.info(f"Session request: {session_request.to_json()}")

        try:
            response = self.send_request(
                "POST",
                urljoin(self._remote_address, Endpoint.DevelopmentSession.value),
                session_request.to_json(),
            )
        except requests.exceptions.ConnectionError:
            logging.error(
                f"Could not start new session on {self._remote_address}. Is your Agent running?"
            )
            logging.error(
                "You can download the TestProject Agent from https://app.testproject.io/#/agents"
            )
            raise AgentConnectException(
                f"Connection error trying to connect to Agent on {self._remote_address}"
            )

        if not response.passed:
            self.__handle_new_session_error(response)

        # The generic driver returns a partial response, so we have to create some fields ourselves
        try:
            session_id = response.data["sessionId"]
        except KeyError:
            session_id = uuid.uuid4()

        try:
            dialect = response.data["dialect"]
        except KeyError:
            dialect = None

        try:
            capabilities = response.data["capabilities"]
        except KeyError:
            capabilities = {}

        start_session_response = SessionResponse(
            dev_socket_port=response.data["devSocketPort"],
            server_address=response.data["serverAddress"],
            session_id=session_id,
            dialect=dialect,
            capabilities=capabilities,
        )
        return start_session_response

    def send_request(self, method, path, body=None) -> OperationResult:
        """Sends HTTP request to Agent

        Args:
            method (str): HTTP method (GET, POST, ...)
            path (str): Relative API route path
            body (dict): Request body

        Returns:
            OperationResult: contains result of the sent request
        """
        with requests.Session() as session:
            if method == "GET":
                response = session.get(path, headers={"Authorization": self._token})
            elif method == "POST":
                response = session.post(
                    path, headers={"Authorization": self._token}, json=body
                )
            elif method == "DELETE":
                response = session.delete(path, headers={"Authorization": self._token})
            else:
                raise SdkException(
                    f"Unsupported HTTP method {method} in send_request()"
                )

        try:
            response.raise_for_status()
            try:
                # For some successful calls, the response body will be empty
                # Parsing it results in a ValueError, so we should handle this
                response_json = response.json()
            except ValueError:
                response_json = {}
            return OperationResult(True, response.status_code, "", response_json)
        except HTTPError as http_error:
            return OperationResult(False, response.status_code, str(http_error), None)

    def send_action_execution_request(
        self, codeblock_guid: str, body: dict
    ) -> ActionExecutionResponse:
        """Sends HTTP request to Agent

        Args:
            codeblock_guid (str): The codeblock GUID to be executed
            body (dict): Parameters to be passed to the Agent

        Returns:
            ActionExecutionResponse: contains result of the sent execution request
        """

        response = self.send_request(
            "POST",
            urljoin(
                urljoin(self._remote_address, Endpoint.ActionExecution.value),
                codeblock_guid,
            ),
            body,
        )

        if not response.passed:
            result = ExecutionResultType.Failed
        else:
            result = (
                ExecutionResultType.Passed
                if response.data["resultType"] == "Passed"
                else ExecutionResultType.Failed
            )

        result_data = response.data["outputs"] if response.passed else None

        return ActionExecutionResponse(result, response.message, result_data)

    @staticmethod
    def get_agent_version(token: str):
        """Requests the current Agent status

        Args:
            token (str): The developer token used to communicate with the Agent

        Returns:
            AgentStatusResponse: contains the response to the sent Agent status request
        """

        with requests.Session() as session:
            response = session.get(
                urljoin(
                    ConfigHelper.get_agent_service_address(), Endpoint.GetStatus.value
                ),
                headers={"Authorization": token},
            )

        try:
            response.raise_for_status()
            try:
                response_json = response.json()
                agent_version = response_json["tag"]
            except ValueError:
                raise SdkException(
                    "Could not parse Agent status response: no JSON response body present"
                )
            except KeyError:
                raise SdkException(
                    "Could not parse Agent status response: element 'tag' not found in JSON response body"
                )
        except HTTPError:
            raise AgentConnectException(
                f"Agent returned HTTP {response.status_code} when trying to retrieve Agent status"
            )

        return AgentStatusResponse(agent_version)

    def report_driver_command(self, driver_command_report: DriverCommandReport):
        """Sends command report to the Agent

        Args:
            driver_command_report: object containing the driver command to be reported
        """

        queue_item = QueueItem(
            report_as_json=driver_command_report.to_json(),
            url=urljoin(self._remote_address, Endpoint.ReportDriverCommand.value),
            token=self._token,
        )

        self._queue.put(queue_item, block=False)

    def report_step(self, step_report: StepReport):
        """Sends step report to the Agent

        Args:
            step_report (StepReport): object containing the step to be reported
        """

        queue_item = QueueItem(
            report_as_json=step_report.to_json(),
            url=urljoin(self._remote_address, Endpoint.ReportStep.value),
            token=self._token,
        )

        self._queue.put(queue_item, block=False)

    def report_test(self, test_report: CustomTestReport):
        """Sends test report to the Agent

        Args:
            test_report (CustomTestReport): object containing the test to be reported
        """

        queue_item = QueueItem(
            report_as_json=test_report.to_json(),
            url=urljoin(self._remote_address, Endpoint.ReportTest.value),
            token=self._token,
        )

        self._queue.put(queue_item, block=False)

    def stop(self):
        """Send all remaining report items in the queue to TestProject"""
        # Send a stop signal to the thread worker
        self._running = False

        # Send a final, empty, report to the queue to ensure that
        # the 'running' condition is evaluated one last time
        self._queue.put(
            QueueItem(report_as_json=None, url=None, token=self._token), block=False
        )

        # Wait until all items have been reported or timeout passes
        self._reporting_thread.join(timeout=self.REPORTS_QUEUE_TIMEOUT)
        if self._reporting_thread.is_alive():
            # Thread is still alive, so there are unreported items
            logging.warning(
                f"There are {self._queue.qsize()} unreported items in the queue"
            )

        try:
            self._sock.shutdown(socket.SHUT_RDWR)
            self._sock.close()
            logging.info(
                f"Connection to Agent at {self._remote_address} closed successfully"
            )
        except socket.error as msg:
            logging.error(
                f"Failed to close socket connection to Agent at {self._remote_address}: {msg}"
            )
            pass

    def __handle_new_session_error(self, response: OperationResult):
        """ Handles errors occurring on creation of a new session with the Agent

        Args:
            response (OperationResult): response from the Agent
        """
        if response.status_code == 401:
            logging.error(
                "Failed to initialize a session with the Agent - invalid developer token supplied"
            )
            logging.error(
                "Get your developer token from https://app.testproject.io/#/integrations/sdk?lang=Python"
                " and set it in the TP_DEV_TOKEN environment variable"
            )
            raise InvalidTokenException(response.message)
        elif response.status_code == 404:
            try:
                error_message = f"Requested browser '{self._capabilities['browserName']}' could not be found on your system"
            except KeyError:
                error_message = "Requested browser could not be found on your system"
            logging.error(error_message)
            raise MissingBrowserException(error_message)
        elif response.status_code == 406:
            logging.error(
                f"Failed to initialize a session with the Agent - obsolete SDK version {ConfigHelper.get_sdk_version()}"
            )
            raise ObsoleteVersionException(response.message)
        else:
            logging.error("Failed to initialize a session with the Agent")
            raise AgentConnectException(
                f"Agent responded with HTTP status {response.status_code}: [{response.message}]"
            )

    def __report_worker(self):
        """Worker method that is polling the queue for items to report"""
        while self._running or self._queue.qsize() > 0:

            item = self._queue.get()
            if isinstance(item, QueueItem):
                item.send()
            else:
                logging.warning(
                    f"Unknown object of type {type(item)} found on queue, ignoring it.."
                )
            self._queue.task_done()


class QueueItem:
    """Helper class representing an item to be reported

    Args:
        report_as_json (dict): JSON payload representing the item to be reported
        url (str): Agent endpoint the payload should be POSTed to
        token (str): Token used to authenticate with the Agent

    Attributes:
        _report_as_json (dict): JSON payload representing the item to be reported
        _url (str): Agent endpoint the payload should be POSTed to
        _token (str): Token used to authenticate with the Agent
    """

    def __init__(self, report_as_json: dict, url: str, token: str):
        self._report_as_json = report_as_json
        self._url = url
        self._token = token

    def send(self):
        """Send a report item to the Agent"""
        if self._report_as_json is None and self._url is None:
            # Skip empty queue items put in the queue on stop()
            return

        with requests.Session() as session:
            response = session.post(
                self._url,
                headers={"Authorization": self._token},
                json=self._report_as_json,
            )
            try:
                response.raise_for_status()
            except HTTPError:
                logging.error(
                    f"Reporting to TestProject returned an HTTP {response.status_code}"
                )
                logging.error(f"Response from Agent: {response.text}")


@unique
class Endpoint(Enum):
    DevelopmentSession = "/api/development/session"
    ActionExecution = "/api/codeblocks/executions"
    ReportDriverCommand = "/api/development/report/command"
    ReportStep = "/api/development/report/step"
    ReportTest = "/api/development/report/test"
    AddonExecution = "/api/addons/executions"
    GetStatus = "/api/status"



from .actionexecutionresponse import ActionExecutionResponse
from .elementsearchcriteria import ElementSearchCriteria
from .proxydescriptor import ProxyDescriptor

__all__ = ["ActionExecutionResponse", "ElementSearchCriteria", "ProxyDescriptor"]



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging

from importlib_metadata import metadata, PackageNotFoundError
from src.testproject.sdk.exceptions import SdkException


def get_sdk_version() -> str:
    """Returns the current SDK version

    Returns:
        str: The current SDK version read from package metadata or an environment variable
    """

    version = None

    try:
        sdk_metadata = metadata('testproject-python-sdk')
        version = sdk_metadata['Version']
    except PackageNotFoundError:
        # This is OK, it just means that there's no previously installed version available
        pass

    logging.debug(f"Version read from package metadata: {version}")

    if version is None:
        # we're not dealing with an installed package, build uses an environment variable
        version = os.environ.get("TP_SDK_VERSION")
        if version is None:
            raise SdkException("No SDK version definition found in metadata or environment variable")

        logging.debug(f"Version read from environment variable: {version}")

    return version



from .custom_command_executor import CustomCommandExecutor
from .custom_appium_command_executor import CustomAppiumCommandExecutor
from .reporting_command_executor import ReportingCommandExecutor
from .generic_command_executor import GenericCommandExecutor

__all__ = [
    "CustomCommandExecutor",
    "CustomAppiumCommandExecutor",
    "ReportingCommandExecutor",
    "GenericCommandExecutor"
]



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from selenium.webdriver.remote.remote_connection import RemoteConnection
from src.testproject.sdk.internal.agent import AgentClient
from src.testproject.sdk.internal.helpers.reporting_command_executor import (
    ReportingCommandExecutor,
)


class CustomCommandExecutor(RemoteConnection, ReportingCommandExecutor):
    """Extension of the Selenium RemoteConnection (command_executor) class

    Args:
        agent_client (AgentClient): Client used to communicate with the TestProject Agent
        remote_server_addr (str): Remote server (Agent) address
    """

    def __init__(self, agent_client: AgentClient, remote_server_addr: str):
        RemoteConnection.__init__(self, remote_server_addr=remote_server_addr)
        ReportingCommandExecutor.__init__(
            self, agent_client=agent_client, command_executor=self
        )

        self.w3c = agent_client.agent_session.dialect == "W3C"

    def execute(self, command: str, params: dict, skip_reporting: bool = False):
        """Execute a Selenium command

        Args:
            command (str): A string specifying the command to execute
            params (dict): A dictionary of named parameters to send with the command as its JSON payload
            skip_reporting (bool): True if command should not be reported to Agent, False otherwise

        Returns:
            response: Response returned by the Selenium remote webdriver server
        """
        self.update_known_test_name()

        response = super().execute(command=command, params=params)

        result = response.get("value")

        passed = self.is_command_passed(response=response)

        if not skip_reporting:
            self._report_command(command, params, result, passed)

        return response



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SessionResponse:
    """Represent a response message object returned by the Agent when initializing a new session

    Args:
        dev_socket_port (int): The developer socket port
        server_address (str): The Agent address
        session_id (str): Unique identifier for the current development session
        dialect (str): Indicates the WebDriver dialect (W3C or OSS)
        capabilities (dict): Desired session capabilities

    Attributes:
        _dev_socket_port (int): The developer socket port
        _server_address (str): The Agent address
        _session_id (str): Unique identifier for the current development session
        _dialect (str): Indicates the WebDriver dialect (W3C or OSS)
        _capabilities (dict): Desired session capabilities
    """

    def __init__(
        self, dev_socket_port: int, server_address: str, session_id: str, dialect: str, capabilities: dict,
    ):
        self._dev_socket_port = dev_socket_port
        self._server_address = server_address
        self._session_id = session_id
        self._dialect = dialect
        self._capabilities = capabilities

    @property
    def dev_socket_port(self) -> int:
        """Getter for the development socket port number"""
        return self._dev_socket_port

    @property
    def server_address(self) -> str:
        """Getter for the Agent address"""
        return self._server_address

    @property
    def session_id(self) -> str:
        """Getter for the unique session ID"""
        return self._session_id

    @property
    def dialect(self):
        """Getter for the WebDriver dialect in use"""
        return self._dialect

    @property
    def capabilities(self) -> dict:
        """Getter for the driver session capabilities"""
        return self._capabilities



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from packaging import version
from src.testproject.enums import EnvironmentVariable
from src.testproject.helpers import ReportHelper, LoggingHelper, ConfigHelper
from src.testproject.rest import ReportSettings
from src.testproject.rest.messages.agentstatusresponse import AgentStatusResponse
from src.testproject.sdk.exceptions import SdkException, AgentConnectException
from src.testproject.sdk.internal.agent import AgentClient
from src.testproject.sdk.internal.helpers import GenericCommandExecutor
from src.testproject.sdk.internal.reporter import Reporter
from src.testproject.sdk.internal.session import AgentSession


class Generic:
    """Used to create a new generic driver instance

    Args:
        token (str): The developer token used to communicate with the agent
        projectname (str): Project name to report
        jobname (str): Job name to report
        disable_reports (bool): set to True to disable all reporting (no report will be created on TestProject)
    """

    __instance = None

    MIN_GENERIC_DRIVER_SUPPORTED_VERSION = "0.64.40"

    def __init__(
        self,
        token: str = None,
        projectname: str = None,
        jobname: str = None,
        disable_reports: bool = False,
    ):
        if Generic.__instance is not None:
            raise SdkException("A driver session already exists")

        LoggingHelper.configure_logging()

        self._token = token if token is not None else ConfigHelper.get_developer_token()

        agent_status_response: AgentStatusResponse = AgentClient.get_agent_version(
            self._token
        )

        if version.parse(agent_status_response.tag) < version.parse(
            Generic.MIN_GENERIC_DRIVER_SUPPORTED_VERSION
        ):
            raise AgentConnectException(
                f"Your current Agent version {agent_status_response.tag} does not support the Generic driver. "
                f"Please upgrade your Agent to the latest version and try again"
            )
        else:
            logging.info(
                f"Current Agent version {agent_status_response.tag} does support Generic driver"
            )

        self.session_id = None

        if disable_reports:
            # Setting the project and job name to empty strings will cause the Agent to not initialize a report
            self._projectname = ""
            self._jobname = ""
        else:
            self._projectname = (
                projectname
                if projectname is not None
                else ReportHelper.infer_project_name()
            )
            self._jobname = (
                jobname if jobname is not None else ReportHelper.infer_job_name()
            )

        reportsettings = ReportSettings(self._projectname, self._jobname)

        capabilities = {"platform": "ANY"}

        self._agent_client: AgentClient = AgentClient(
            token=self._token, capabilities=capabilities, reportsettings=reportsettings,
        )
        self._agent_session: AgentSession = self._agent_client.agent_session

        self.command_executor = GenericCommandExecutor(agent_client=self._agent_client)

        Generic.__instance = self

    @classmethod
    def instance(cls):
        """Returns the singleton instance of the driver object"""
        return Generic.__instance

    def start_session(self, capabilities, browser_profile=None):
        """Sets capabilities and sessionId obtained from the Agent when creating the original session."""
        self.session_id = self._agent_session.session_id
        logging.info(f"Session ID is {self.session_id}")

    def report(self) -> Reporter:
        """Enables access to the TestProject reporting actions from the driver object"""
        return Reporter(self.command_executor)

    def quit(self):
        """Quits the driver and stops the session with the Agent, cleaning up after itself."""
        # Report any left over driver command reports
        self.command_executor.clear_stash()

        # Report test explicitly as this is not done automatically
        self.command_executor.report_test()

        # Make instance available again
        Generic.__instance = None

        # Stop the Agent client
        self.command_executor.agent_client.stop()

        # Clean up any environment variables set in the decorator
        for env_var in [
            EnvironmentVariable.TP_TEST_NAME,
            EnvironmentVariable.TP_PROJECT_NAME,
            EnvironmentVariable.TP_JOB_NAME,
        ]:
            EnvironmentVariable.remove(env_var)



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import os
import sys
import traceback

from src.testproject.sdk.exceptions import SdkException

from src.testproject.sdk.drivers.webdriver import Remote, Generic

from src.testproject.sdk.drivers.webdriver.base import BaseDriver


def report_assertion_errors(func):
    """Enables automatic logging of failed assertions, including a screenshot"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except AssertionError as ae:
            _, _, tb = sys.exc_info()
            tb_info = traceback.extract_tb(tb)

            # Depending on the unit testing framework in use we need a different line in the stack
            line_index = -1 if os.getenv("PYTEST_CURRENT_TEST") is not None else -2

            # Depending on the unit testing framework we format the reported error differently
            description = (
                ae.__repr__()
                if os.getenv("PYTEST_CURRENT_TEST") is not None
                else str(ae)
            )

            _, line, function, text = tb_info[line_index]
            driver = __get_active_driver_instance()
            driver.report().step(
                description=description,
                message=f"Assertion failed on line {line} in {function}",
                passed=False,
                screenshot=False,
            )
            # raise the AssertionError again to make the test execution fail as intended
            raise ae

    return wrapper


def __get_active_driver_instance():
    """Get the current driver instance in use (BaseDriver, Remote or Generic) """
    driver = BaseDriver.instance()
    if driver is None:
        driver = Remote.instance()
        if driver is None:
            driver = Generic.instance()
            if driver is None:
                raise SdkException(
                    "No active driver instance found, so cannot report failed assertion"
                )

    return driver



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from src.testproject.enums import ExecutionResultType
from src.testproject.sdk.drivers.actions import DriverActions
from src.testproject.sdk.internal.agent import AgentClient
from selenium.webdriver.common.by import By
from src.testproject.sdk.drivers.actions.action_guids import web_actions


class WebActions(DriverActions):
    """Offers methods to execute web actions

    Args:
        agent_client (AgentClient): client to communicate with the Agent
        timeout (int): timeout for action execution
    """

    def __init__(self, agent_client: AgentClient, timeout: int):
        super().__init__(agent_client, timeout)

    def move_mouse_to_element(self, by: By, by_value: str) -> bool:
        """Moves the mouse to the middle of an element and scrolls it into view

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        response = self.action_execute(web_actions["MOVE_MOUSE_TO_ELEMENT_ID"], {}, by, by_value)
        return response.executionresulttype == ExecutionResultType.Passed

    def navigate_forward(self) -> bool:
        """Navigate forward

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        response = self.action_execute(web_actions["NAVIGATE_FORWARD_ID"], {}, None, "")
        return response.executionresulttype == ExecutionResultType.Passed

    def navigate_back(self) -> bool:
        """Navigate back

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        response = self.action_execute(web_actions["NAVIGATE_BACK_ID"], {}, None, "")
        return response.executionresulttype == ExecutionResultType.Passed

    def refresh(self) -> bool:
        """Refresh the current browser tab

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        response = self.action_execute(web_actions["REFRESH_ID"], {}, None, "")
        return response.executionresulttype == ExecutionResultType.Passed

    def navigate_to_url(self, url: str) -> bool:
        """Navigates to the specified URL in the active browser tab

        Args:
            url (str): The URL to navigate to

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        body = {"url": url}
        response = self.action_execute(web_actions["NAVIGATE_TO_URL_ID"], body, None, "")
        return response.executionresulttype == ExecutionResultType.Passed

    def get_current_url(self) -> str:
        """Retrieves the current URL from the active browser tab

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        response = self.action_execute(web_actions["GET_CURRENT_URL_ID"], {}, None, "")
        if response.executionresulttype != ExecutionResultType.Passed:
            return None
        return response.outputs["url"]

    def scroll_window(self, pixels_x_axis: int, pixels_y_axis: int) -> bool:
        """Navigates to the specified URL in the active browser tab

        Args:
            pixels_x_axis (int): Amount of pixels to scroll on X axis. A negative value means opposite direction
            pixels_y_axis (int): Amount of pixels to scroll on Y axis. A negative value means opposite direction

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        body = {"x": pixels_x_axis, "y": pixels_y_axis}
        response = self.action_execute(web_actions["SCROLL_WINDOW_ID"], body, None, "")
        return response.executionresulttype == ExecutionResultType.Passed

    def select_all_options_by_value(self, by: By, by_value: str, option_value: str) -> bool:
        """Select all options that have a value matching the argument

        Args:
            by (By): Selenium locator strategy (By.ID, By.NAME, ...)
            by_value (str): The associated value for the locator strategy
            option_value (str): The value of the option to select

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        body = {"value": option_value}
        response = self.action_execute(web_actions["SELECT_ALL_OPTIONS_BY_VALUE_ID"], body, by, by_value,)
        return response.executionresulttype == ExecutionResultType.Passed

    def switch_to_window(self, index: int) -> bool:
        """Select a window using its index

        Args:
            index (int): The index of the window to select

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        body = {"index": index}
        response = self.action_execute(web_actions["SWITCH_TO_WINDOW_ID"], body, None, "")
        return response.executionresulttype == ExecutionResultType.Passed

    def close_window(self, index: int) -> bool:
        """Closes the window with the given index

        Args:
            index (int): The index of the window to close

        Returns:
            bool: True if action was performed successfully, False otherwise
        """
        body = {"index": index}
        response = self.action_execute(web_actions["CLOSE_WINDOW_ID"], body, None, "")
        return response.executionresulttype == ExecutionResultType.Passed



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from enum import Enum, unique


@unique
class EnvironmentVariable(Enum):
    """Enumeration of environment variable names used in the SDK"""

    TP_TEST_NAME = "TP_TEST_NAME"
    TP_PROJECT_NAME = "TP_PROJECT_NAME"
    TP_JOB_NAME = "TP_JOB_NAME"

    def remove(self):
        """Try and remove the environment variable, proceed if the variable doesn't exist"""
        try:
            os.environ.pop(self.value)
        except KeyError:
            pass



import setuptools

from src.testproject import definitions

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="testproject-python-sdk",
    version=definitions.get_sdk_version(),
    author="TestProject",
    author_email="contact@testproject.io",
    description="Selenium and Appium powered SDK for TestProject.io",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://testproject.io/selenium-appium-powered-sdk/",
    packages=setuptools.find_packages(exclude=['tests', 'tests.*', 'proxy_examples', 'proxy_examples.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=[
        "selenium==3.141.0",
        "Appium-Python-Client>=1.0.1",
        "decorator>=4.4.2",
        "requests>=2.24.0",
        "importlib-metadata>=1.7.0",
        "packaging>=20.4"
    ],
)



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class ReportSettings:
    """Contains settings to be used in the report.

        Args:
            projectname (str): Project name to report
            jobname (str): Job name to report

        Attributes:
            _projectname (str): Project name to report
            _jobname (str): Job name to report
    """

    def __init__(self, projectname: str, jobname: str):
        self._projectname = projectname
        self._jobname = jobname

    @property
    def projectname(self):
        """Getter for the project name"""
        return self._projectname

    @property
    def jobname(self):
        """Getter for the job name"""
        return self._jobname



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class DriverCommandReport:
    """Payload object sent to the Agent when reporting a driver command.

    Args:
        command (str): The name of the command that was executed
        command_params (dict): Parameters associated with the command
        result (dict): The result of the command that was executed
        passed (bool): Indication whether or not command execution was performed successfully
        screenshot (str): Screenshot as base64 encoded string

    Attributes:
        _command (str): The name of the command that was executed
        _command_params (dict): Parameters associated with the command
        _result (dict): The result of the command that was executed
        _passed (bool): Indication whether or not command execution was performed successfully
        _screenshot (str): Screenshot as base64 encoded string
    """

    def __init__(
        self,
        command: str,
        command_params: dict,
        result: dict,
        passed: bool,
        screenshot: str = None,
    ):
        self._command = command
        self._command_params = command_params
        self._result = result
        self._passed = passed
        self._screenshot = screenshot

    @property
    def command(self) -> str:
        """Getter for the command property"""
        return self._command

    @property
    def command_params(self) -> dict:
        """Getter for the command_params property"""
        return self._command_params

    @property
    def result(self) -> dict:
        """Getter for the result property"""
        return self._result

    @property
    def passed(self) -> bool:
        """Getter for the passed property"""
        return self._passed

    @property
    def screenshot(self) -> str:
        """Getter for the screenshot property"""
        return self._screenshot

    @screenshot.setter
    def screenshot(self, value: str):
        """Setter for the screenshot property"""
        self._screenshot = value

    def to_json(self):
        """Creates a JSON representation of the current DriverCommandReport instance

            Returns:
                dict: JSON representation of the current instance
        """
        payload = {
            "commandName": self._command,
            "commandParameters": self._command_params,
            "result": self._result,
            "passed": self._passed,
        }

        # Add screenshot to report if it is provided
        if self._screenshot is not None:
            payload["screenshot"] = self._screenshot

        return payload

    def __eq__(self, other):
        """Custom equality function, used in report stashing"""
        if not isinstance(other, DriverCommandReport):
            return NotImplemented

        return (
            self._command == other._command
            and self._command_params == other._command_params
            and self._result == other._result
            and self._passed == other._passed
        )

    def __hash__(self):
        """Implement hash to allow objects to be used in sets and dicts"""
        return hash(
            (
                self.command,
                self.command_params,
                self.result,
                self.passed,
                self.screenshot,
            )
        )



from .sdkexception import SdkException
from .agentconnectexception import AgentConnectException
from .invalidtokenexception import InvalidTokenException
from .obsoleteversionexception import ObsoleteVersionException
from .missingbrowserexception import MissingBrowserException

__all__ = [
    "SdkException",
    "AgentConnectException",
    "InvalidTokenException",
    "ObsoleteVersionException",
    "MissingBrowserException",
]



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import inspect

from selenium.webdriver.remote.command import Command
from src.testproject.helpers import ReportHelper
from src.testproject.rest.messages import DriverCommandReport, CustomTestReport
from src.testproject.sdk.internal.agent import AgentClient
from src.testproject.sdk.internal.helpers.redact_helper import RedactHelper


class ReportingCommandExecutor:
    """Class responsible for executing commands and reporting them

    Args:
        agent_client (AgentClient): Client used to communicate with the TestProject Agent
        command_executor: The command executor used to send WebDriver commands (Selenium or Appium)

    Attributes:
        _agent_client (AgentClient): Client used to communicate with the TestProject Agent
        _command_executor: The command executor used to send WebDriver commands (Selenium or Appium)
        _disable_reports (bool): True if all reporting is disabled, False otherwise
        _disable_auto_test_reports (bool): True if automatic reporting of tests is disabled, False otherwise
        _disable_command_reports (bool): True if driver command reporting is disabled, False otherwise
        _disable_redaction (bool): True if reporting steps should be redacted, False otherwise
        _stashed_command (DriverCommandReport): contains stashed driver command for preventing duplicates
        inside WebDriverWait
        _latest_known_test_name (str): contains latest known test name
        _excluded_test_names (list): contains a list of test names that should not be reported
    """

    def __init__(self, agent_client: AgentClient, command_executor):
        self._agent_client = agent_client
        self._command_executor = command_executor
        self._disable_reports = False
        self._disable_auto_test_reports = False
        self._disable_command_reports = False
        self._disable_redaction = False
        self._stashed_command = None
        self._latest_known_test_name = ReportHelper.infer_test_name()
        self._excluded_test_names = list()

    @property
    def disable_reports(self) -> bool:
        """Getter for the disable_reports flag"""
        return self._disable_reports

    @disable_reports.setter
    def disable_reports(self, value: bool):
        """Setter for the disable_reports flag"""
        self._disable_reports = value

    @property
    def disable_auto_test_reports(self) -> bool:
        """Getter for the disable_auto_test_reports flag"""
        return self._disable_auto_test_reports

    @disable_auto_test_reports.setter
    def disable_auto_test_reports(self, value: bool):
        """Setter for the disable_auto_test_reports flag"""
        self._disable_auto_test_reports = value

    @property
    def disable_command_reports(self) -> bool:
        """Getter for the disable_command_reports flag"""
        return self._disable_command_reports

    @disable_command_reports.setter
    def disable_command_reports(self, value: bool):
        """Setter for the disable_command_reports flag"""
        self._disable_command_reports = value

    @property
    def disable_redaction(self) -> bool:
        """Getter for the disable_redaction flag"""
        return self._disable_redaction

    @disable_redaction.setter
    def disable_redaction(self, value: bool):
        """Setter for the disable_redaction flag"""
        self._disable_redaction = value

    @property
    def excluded_test_names(self) -> list:
        """Getter for the list of excluded test names"""
        return self._excluded_test_names

    @excluded_test_names.setter
    def excluded_test_names(self, value: list):
        """Setter for the list of excluded test names"""
        self._excluded_test_names = value

    @property
    def agent_client(self):
        """Getter for the Agent client associated with this connection"""
        return self._agent_client

    def _report_command(self, command: str, params: dict, result: dict, passed: bool):
        """Reports a driver command to the TestProject platform

        Args:
            command (str): The driver command to execute
            params (dict): Named parameters to send with the command as its JSON payload
            result (dict): The response returned by the Selenium remote webdriver server
            passed (bool): True if the command execution was successful, False otherwise
        """
        if command == Command.QUIT:
            if not self.disable_auto_test_reports:
                self.report_test()
            return  # This ensures that the actual driver.quit() command is not included in the report

        if not self._disable_redaction:
            params = RedactHelper(self).redact_command(command, params)

        # If the command is executed as part of a wait loop, we don't want to report it every time
        self._is_webdriverwait = False

        # See if the command is executed inside a wait loop
        for frame in inspect.stack().__reversed__():
            if str(frame.filename).find("wait.py") > 0:
                self._is_webdriverwait = True
                break

        driver_command_report = DriverCommandReport(command, params, result, passed)

        if not passed:
            driver_command_report.screenshot = self.create_screenshot()

        if self._is_webdriverwait:
            if not self._disable_reports and not self.disable_command_reports:
                # Only stash the command for reporting later when driver command reporting is enabled
                self._stashed_command = driver_command_report
            return  # Do not report the command right away

        if not self._disable_reports and not self.disable_command_reports:
            if self._stashed_command is not None:
                # report the stashed command and clear it
                self.agent_client.report_driver_command(self._stashed_command)
                self._stashed_command = None
            # report the current command
            self.agent_client.report_driver_command(driver_command_report)

    def update_known_test_name(self):
        """Infers the current test name and if different from the latest known test name, reports a test"""
        current_test_name = ReportHelper.infer_test_name()

        # Actions inside a unittest tearDown or tearDownClass method should be reported as part of the test
        in_unittest_teardown = ReportHelper.find_unittest_teardown()

        if (
            current_test_name not in [self._latest_known_test_name, "Unnamed Test"]
            and not in_unittest_teardown
        ):
            # the name of the test method has changed and we're not inside a unittest teardown method,
            # so we need to report a test
            if not self.disable_auto_test_reports:
                self.report_test()
            # update the latest known test name for future reports
            self._latest_known_test_name = current_test_name

    def report_test(self):
        """Sends a test report to the Agent if this option is not explicitly disabled
        """

        if not self._latest_known_test_name == "Unnamed Test":

            # only report those tests that have been identified as one when their names were inferred
            if self._disable_reports:
                # test reporting has been disabled by the user
                logging.debug(f"Test [{self._latest_known_test_name}] - [Passed]")
                return

            if self._latest_known_test_name in self._excluded_test_names:
                # test has been marked as 'to be excluded, so do not report it
                logging.debug(
                    f"Test [{self._latest_known_test_name}] - Reporting skipped (marked as 'To be excluded')"
                )
                return

            custom_test_report = CustomTestReport(
                name=self._latest_known_test_name, passed=True
            )
            self.agent_client.report_test(custom_test_report)

    def create_screenshot(self) -> str:
        """Creates a screenshot (PNG) and returns it as a base64 encoded string

        Returns:
            str: The base64 encoded screenshot in PNG format (or None if screenshot taking fails)
        """
        create_screenshot_params = {
            "sessionId": self.agent_client.agent_session.session_id
        }
        create_screenshot_response = self._command_executor.execute(
            Command.SCREENSHOT, create_screenshot_params, True
        )
        try:
            return create_screenshot_response["value"]
        except KeyError as ke:
            logging.error(f"Error occurred creating a screenshot: {ke}")
            logging.error(
                f"Response from RemoteWebDriver: {create_screenshot_response}"
            )
            return None

    def clear_stash(self):
        """Reports stashed command if there is one left. Should be called when session ends to prevent
           wait-related commands from not being reported.
        """
        if not self._disable_reports and not self.disable_command_reports:
            if self._stashed_command is not None:
                # report the stashed command and clear it
                self.agent_client.report_driver_command(self._stashed_command)
                self._stashed_command = None

    def is_command_passed(self, response: dict) -> bool:
        """Determine command result based on response using state and status.

        Args:
            response (dict): The response returned by the Selenium remote webdriver server

        Returns:
            bool: True if passed, otherwise False.
        """
        # Both None and 0 response status values indicate command execution was OK
        return True if response.get("status") in [None, 0] else False



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from src.testproject.enums import ExecutionResultType


class AddonExecutionResponse:
    """Object representing the response returned when execution a custom action

    Attributes:
        _executionresulttype (ExecutionResultType): The result of the action execution (Passed / Failed / ...)
        _message (str): A message returned by the agent in response to the action execution
        _fields (list): A potentially updated list of fields
    """

    def __init__(self):
        self._executionresulttype = None
        self._message = None
        self._fields = None

    @property
    def executionresulttype(self) -> ExecutionResultType:
        """Getter for the execution result type"""
        return self._executionresulttype

    @executionresulttype.setter
    def executionresulttype(self, value: ExecutionResultType):
        """Setter for the execution result type"""
        self._executionresulttype = value

    @property
    def message(self) -> str:
        """Getter for the action execution response message"""
        return self._message

    @message.setter
    def message(self, value: str):
        """Setter for the action execution response message"""
        self._message = value

    @property
    def fields(self) -> list:
        """Getter for the list of action execution response fields"""
        return self._fields

    @fields.setter
    def fields(self, value: list):
        """Setter for the list of action execution response fields"""
        self._fields = value



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from src.testproject.enums import FindByType


class ElementSearchCriteria:
    """Defines an object type representing search criteria for a web element

    Args:
        find_by_type (FindByType): The locator strategy to be used (id, name, etc.)
        by_value (str): The associated locator strategy value
        index (int): An index indicating which occurrence of the element should be used

    Attributes:
        _find_by_type (FindByType): The locator strategy to be used (id, name, etc.)
        _by_value (str): The associated locator strategy value
        _index (int): An index indicating which occurrence of the element should be used
    """
    def __init__(self, find_by_type: FindByType, by_value: str, index: int = -1):
        self._find_by_type = find_by_type
        self._by_value = by_value
        self._index = index

    @property
    def find_by_type(self) -> FindByType:
        """Getter for the element locator strategy type"""
        return self._find_by_type

    @find_by_type.setter
    def find_by_type(self, value: FindByType):
        """Setter for the element locator strategy type"""
        self._find_by_type = value

    @property
    def by_value(self) -> str:
        """Getter for the element locator strategy value"""
        return self._by_value

    @by_value.setter
    def by_value(self, value: str):
        """Setter for the element locator strategy value"""
        self._by_value = value

    @property
    def index(self) -> int:
        """Getter for the element locator index"""
        return self._index

    @index.setter
    def index(self, value: int):
        """Setter for the element locator index"""
        self._index = value

    def to_json(self):
        """Returns a JSON representation of the object to be sent to the Agent"""
        return {
            "byType": self._find_by_type.name,
            "byValue": self._by_value,
            "index": self._index,
        }



# Copyright 2020 TestProject (https://testproject.io)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import socket

from urllib.parse import urlparse
from src.testproject.sdk.exceptions import AgentConnectException


class SocketHelper:
    """Contains helper methods for interacting with the Agent development socket"""

    @staticmethod
    def create_connection(socket_address: str, socket_port: int) -> socket:
        """Parses the agent service address and attempts to create a socket connection

            Args:
                socket_address (str): The address for the socket
                socket_port (int): The development socket port to connect to

            Returns:
                socket: Socket object that has been created and connected to
        """
        host = urlparse(socket_address).hostname

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        sock.connect((host, socket_port))

        if not SocketHelper.is_socket_connected(sock):
            raise AgentConnectException("Failed connecting to Agent socket")

        logging.info(f"Socket connection to {host}:{socket_port} established successfully")

        return sock

    @staticmethod
    def is_socket_connected(sock) -> bool:
        """Sends a simple message to the socket to see if it's connected

            Args:
                sock (socket): The socket object

            Returns:
                bool: True if the socket is connected, False otherwise
        """
        try:
            sock.send("test".encode("utf-8"))
            return True
        except socket.error as msg:
            logging.error(f"Socket not connected: {msg}")
            return False
